/*
 * General functionality supporting file paging.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "file_pager.h"
#include "file_paging.h"

#include <systypes/fcntl.h>



/* The file paging coordinator provides a memory page collection to the page
   mappers it creates. */

FilePaging::FilePaging(Pages *pages)
: _pages(pages)
{
}



/* Return any registered provider for the given 'fileid' or NULL if no such
   provider is registered. */

Accountable *FilePaging::get(fileid_t fileid)
{
    FileMapping::iterator entry = _providers.find(fileid);
    Accountable *provider;

    if (entry == _providers.end())
        provider = NULL;
    else
        provider = entry->second;

    return provider;
}

/* Remove a provider and its resources for the given 'fileid'. */

void FilePaging::remove(fileid_t fileid, Accountable *provider)
{
    _providers.erase(fileid);
    delete provider;
}

/* Register a 'provider' for the given 'fileid'. */

void FilePaging::set(fileid_t fileid, Accountable *provider)
{
    FileMapping::iterator entry = _providers.find(fileid);

    if (entry != _providers.end())
        return;

    _providers[fileid] = provider;
}



/* Convert opening flags to map-compatible paging flags. */

map_flags_t FilePaging::get_flags(flags_t flags)
{
    return flags & (O_WRONLY | O_RDWR) ? L4RE_DS_MAP_FLAG_RW : L4RE_DS_MAP_FLAG_RO;
}



/* Obtain a provider for the 'fileid' or register a new one in the
   paging object. */

long FilePaging::get_provider(FileOpening *opening, const char *path, flags_t flags, fileid_t fileid, FileProvider **file_provider)
{
    /* Obtain any registered provider. */

    Accountable *provider = get(fileid);

    if (provider != NULL)
    {
        /* Distinguish between file providers and other objects that may be
           registered. For files specifically, a file provider is needed
           because it will provide a page mapper. */

        *file_provider = dynamic_cast<FileProvider *>(provider);

        if ((*file_provider) != NULL)
            return L4_EOK;
        else
            return -L4_EIO;
    }

    /* Make an accessor, page mapper, and a provider to encapsulate them. */

    Accessor *accessor;
    long err = opening->make_accessor(path, flags, fileid, &accessor);

    if (err)
        return err;

    PageMapper *mapper = new PageMapper(accessor, _pages);
    *file_provider = new FileProvider(mapper);

    /* Register the provider. */

    set(fileid, *file_provider);
    return L4_EOK;
}



/* Return a pager initialised with a provider, page mapper and accessor. */

long FilePaging::get_pager(FileOpening *opening, const char *path, flags_t flags, fileid_t fileid, Pager **pager)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Obtain any existing provider registered for the file, or make a new
       provider. */

    FileProvider *provider;
    long err = get_provider(opening, path, flags, fileid, &provider);

    if (err)
        return err;

    /* Initialise the pager with the provider and a reference to this object for
       closing the provider, mapper and accessor. */

    provider->attach();
    *pager = new FilePager(fileid, provider, get_flags(flags), this);
    return L4_EOK;
}

/* Detach a pager, potentially removing its resources. */

void FilePaging::detach_pager(fileid_t fileid, Accountable *provider)
{
    std::lock_guard<std::mutex> guard(_lock);

    if (!provider->detach())
        remove(fileid, provider);
}

// vim: tabstop=4 expandtab shiftwidth=4
